
import secrets
import time
from functools import wraps
from flask_httpauth import HTTPAuth
from werkzeug.security import safe_str_cmp


class Authentication(HTTPAuth):

    def __init__(self, users, token_timeout_seconds=6000, token_size=32):
        self._users = users
        self._tokens = {}
        self._token_timeout_seconds = token_timeout_seconds
        self._token_size = token_size
        if self._token_size < 4:
            raise(ValueError('Token size must be >= 4'))
        HTTPAuth.__init__(self, scheme='Token')

    def get_auth_password(self, auth):
        password = None
        if auth and auth.username:
            password = self._users[auth.username]
        return password

    def init_app(self, app):
        app.config['SECRET_KEY'] = secrets.token_hex(64)

    def create_token(self, username):
        token = secrets.token_hex(self._token_size)
        token_data = {
            'timeout': time.time() + self._token_timeout_seconds,
            'username': username
        }
        self._tokens[token] = token_data
        return token

    def revoke_token(self, username, token):
        is_found = False
        new_tokens = {}
        for token_item, token_data in self._tokens.items():
            if token_item == token:
                is_found = True
            elif token_data['timeout'] > time.time():
                new_tokens[token_item] = token_data
        self._tokens = new_tokens
        return is_found

    def get_tokens(self, username):
        token_list = []
        for token_item, token_data in self._tokens.items():
            new_tokens = {}
            if token_data['timeout'] > time.time():
                new_tokens[token_item] = token_data
                if token_data['username'] == username:
                    token_list.append(token_item)
            self._tokens = new_tokens
        return token_list

    def is_basic_authorized(self, auth):
        auth_password = auth.password
        password = self.get_auth_password(auth)
        return safe_str_cmp(auth_password, password)

    def is_token_valid(self, token):
        if token and token in self._tokens:
            token_data = self._tokens[token]
            if token_data['timeout'] > time.time():
                return True
        return False

    def authenticate(self, auth, password):
        token = None
        if auth:
            token = auth['token']
        return self.is_token_valid(token)

    def login_maybe_required(self, f):
        @wraps(f)
        def decorated(*args, **kwargs):
            if self._users:
                return super(Authentication, self).login_required(f)(*args, **kwargs)
            return f(*args, **kwargs)
        return decorated

    @property
    def is_enabled(self):
        return self._users
