import logging
from flask import Flask
from flask_cors import CORS
from flask_swagger_ui import get_swaggerui_blueprint

from invokable.authentication import Authentication
from invokable.constants import API_VERSION
from invokable.invoke_handler import InvokeHandler
from invokable.storage import storage_load
from invokable.ui import (
    get_agent_blueprint,
    get_authentication_blueprint,
    get_invoke_blueprint,
    get_metadata_blueprint,
    get_storage_blueprint,
    get_swagger_blueprint
)

from invokable.utils import (
    generate_url,
    load_route_paths
)

logger = logging.getLogger('invokable-py.invokable_server')


class InvokableServer:
    """

    Create a web server that handles invoke services.

    :param str host:
        Host name of the server.

        Default: **127.0.0.1**

    :param int port: Port number of the server.

        Default: **80**

    :param list,str package: Package name or list of package names to register.

        Default: **None**

    :param list,str module:
        Module name or list of module names to register.

        Default: **None**

    :param invoke_handler:
        Invoke handler to use.

        Default: **None**

    :type invoke_handler: :class:`.InvokeHandler`

    :param bool is_swagger:

        If True show swagger api docs for
        all invokable services, else no swagger documentation.

        Default: **True**

    :param str title:
        Title of the web server for swagger documentation.

    :param str description:
        Description of the web server for swagger documentation.

    :param str api_version:
        API Version of the web server for swagger documentation.

    :param str,list swagger_list:
        list or a str name or
        address of the invokable services to show in swagger,
        if == [] then only show invokable api calls.

        Default: **None**

    :param str,list swagger_list_type:
        list or a str with 'sync' and or 'async'.

        Default: **None**

    :param dict route_paths:
        Server route paths to the different api endpoints.
        This can be a single dict name, value or multiple name/values.

        Default: See :py:mod:`invokable.constants`

    :param string ddo_text:
        The agent server ddo record in text format. This is used to provide
        the agent API. If you do not pass a valid DDO for this service,
        then the api calls api/getDDO and api/status will not be provided.

    :param dict users:
        Dict set of usernames and their passwords.

    :param string, dict storage:
        If you wish to use the metadata and asset storage on this agent/server.
            Then you can pass a string in the form

            + **memory** - memory storage for both metadata and asset.

            + **sqlite:<filename>** - sqlite storage. <filename> is the name of the sqlite db file.

            You can also pass an object that is inherited from the :class:`.Storage`
            to define your own storage options.

            So setting storage to 'memory' is the same as setting storage to StorageMemory(),
            and 'sqlite:/tmp/test.db' is the same as passing the object StorageSqlite('/tmp/test.db')

    :param dict data:
        Invoke data to pass to the invoke run methods.

    .. code-block:: python

        # Only show the two sync invokable services in the swagger docs

        from invokable import InvokableServer
        server = InvokableServer(
            swagger_list = ['service_1', 'service_2'],
            swagger_list_type = 'sync',
            package = 'myservices.service'
        )
        app = server.create_app()
        app.run(host=server.host, port=server.port)

    """

    def __init__(self, host='127.0.0.1', port=80, package=None, module=None, invoke_handler=None,
                 is_swagger=True, title='Invokable Server Title', description='Invokable Server Description',
                 api_version=None, swagger_list=None, swagger_list_type=None, route_paths=None,
                 ddo_text=None, users=None, storage=None, data=None):

        self._host = host
        self._port = port
        self._title = title
        self._description = description
        self._api_version = api_version
        if self._api_version is None:
            self._api_version = API_VERSION
        self._is_swagger = is_swagger
        self._swagger_list = swagger_list
        self._swagger_list_type = swagger_list_type
        self._invoke_handler = invoke_handler
        if self._invoke_handler is None:
            self._invoke_handler = InvokeHandler()
        self._app = None
        if package:
            self.add_package(package)
        if module:
            self.add_module(module)

        # load the route paths
        self._route_paths = load_route_paths(route_paths)

        self._ddo_text = ddo_text
        if self._ddo_text and not isinstance(self._ddo_text, str):
            raise TypeError(f'You must provide the DDO as a text string')

        self._authentication = Authentication(users)

        self._storage = storage_load(storage)

        self._data = data

    def create_app(self, flask_name=None, flask_kwargs={}, app=None):
        """

        Create a new Flask app, with the correct Flask blueprints setup.

        :param object app: Optional Flask app to pass instead of creating a new one.
        :param list args: Arguments to pass to the invoke service
        :param dict kwargs: Dict arguments to pass to the invoke service

        :return: A Flask app.
        :type: Object

        .. code-block:: python

            # A simple example

            from invokable import InvokableServer
            server = InvokableServer(package = 'myservices.service')
            app = server.create_app()
            app.run(host=server.host, port=server.port)

        """

        self._app = app
        if self._app is None:
            if flask_name is None:
                flask_name = __name__
            self._app = Flask(flask_name, **flask_kwargs)

        CORS(self._app)
        self._app.config['server'] = self
        self._authentication.init_app(self._app)

        self.register_metadata_api(
            self._app,
            self._route_paths['meta'],
            self._invoke_handler,
            self._storage,
            self._authentication
        )
        self.register_invoke_api(
            self._app,
            self._route_paths['invoke'],
            self._invoke_handler,
            self._authentication,
            self._data
        )

        if self._ddo_text:
            self.register_agent_api(
                self._app,
                self._route_paths['agent'],
                self._title, self._description,
                self._api_version, self._ddo_text
            )

        if self._authentication.is_enabled:
            self.register_authentication_api(
                self._app,
                self._route_paths['auth'],
                self._authentication
            )

        if self._storage:
            self.register_storage_api(
                self._app,
                self._route_paths['storage'],
                self._storage,
                self._authentication
            )

        if self._is_swagger:
            self.register_swagger_api(
                self._app,
                self._route_paths['swagger_api'],
                self._route_paths['swagger_docs'],
                self._route_paths['invoke'],
                self._invoke_handler,
                self._title,
                self._description,
                self._api_version,
                self._swagger_list,
                self._swagger_list_type
            )
        return self._app

    def register_agent_api(self, app, uri, title, description, api_version, ddo_text):
        """

        Register the agent api as a Flask blueprint.

        This method is called by the :func:`create_app` method, so you only need
        to call this method if you do not want to call the :func:`create_app` method.

        :param Object app: Flask app object.
        :param str uri: URI of the api
        :param str title: Title of the agent
        :param str description: Descritpion of the agent
        :param str api_version: Version of the agent api
        :param object ddo: DDO object to that has been registered on the network

        """
        blueprint = get_agent_blueprint(uri, title, description, api_version, ddo_text)
        app.register_blueprint(blueprint, url_prefix=uri)

    def register_authentication_api(self, app, uri, authentication):
        """

        Register the authentication api as a Flask blueprint.

        This method is called by the :func:`create_app` method, so you only need
        to call this method if you do not want to call the :func:`create_app` method.

        :param Object app: Flask app object.
        :param str uri: URI of the api
        :param object authentication: Authentication object to do token and basic auth

        """
        blueprint = get_authentication_blueprint(uri, authentication)
        app.register_blueprint(blueprint, url_prefix=uri)

    def register_metadata_api(self, app, uri, invoke_handler, storage, authentication):
        """

        Register the metadata api as a Flask blueprint.

        This method is called by the :func:`create_app` method, so you only need
        to call this method if you do not want to call the :func:`create_app` method.

        :param Object app: Flask app object.
        :param str uri: URI of the api
        :param invoke_handler: Invoke handler to use.
        :type invoke_handler: :class:`.InvokeHandler`
        :param storage: Storage object to save use metadata, this can be None for no user storage.
        :type storage: :class:`.Storage`

        """
        blueprint = get_metadata_blueprint(uri, invoke_handler, storage, authentication)
        app.register_blueprint(blueprint, url_prefix=uri)

    def register_storage_api(self, app, uri, storage, authentication):
        """

        Register the metadata api as a Flask blueprint.

        This method is called by the :func:`create_app` method, so you only need
        to call this method if you do not want to call the :func:`create_app` method.

        :param Object app: Flask app object.
        :param str uri: URI of the api
        :param invoke_handler: Invoke handler to use.
        :type invoke_handler: :class:`.InvokeHandler`
        :param storage_metadata: Storage object to save user asset storage, this can be None for no user storage.
        :type storage: :class:`.Storage`

        """
        blueprint = get_storage_blueprint(uri, storage, authentication)
        app.register_blueprint(blueprint, url_prefix=uri)

    def register_invoke_api(self, app, uri, invoke_handler, authentication, data=None):
        """

        Register the invoke api as a Flask blueprint.

        This method is called by the :func:`create_app` method, so you only need
        to call this method if you do not want to call the :func:`create_app` method.

        :param Object app: Flask app object.
        :param str uri: URI of the api
        :param invoke_handler: Invoke handler to use.
        :type invoke_handler: :class:`.InvokeHandler`
        :param dict data: Dict of user data to pass to the invokable service

        """
        blueprint = get_invoke_blueprint(uri, invoke_handler, authentication, data=data)
        app.register_blueprint(blueprint, url_prefix=uri)

    def register_swagger_api(self, app, api_uri, doc_uri, invoke_uri, invoke_handler,
                             title, description, api_version, swagger_list, swagger_list_type):
        """

        Register the swagger api as a Flask blueprint.

        This method is called by the :func:`create_app` method, so you only need
        to call this method if you do not want to call the :func:`create_app` method.

        :param Object app: Flask app object.
        :param str api_uri: Swagger api URI.
        :param str doc_uri: Swagger document URI.
        :param str invoke_uri: Inovke URI.
        :param invoke_handler: Invoke handler to use.
        :type invoke_handler: :class:`.InvokeHandler`
        :param str title: Title of the web server for swagger documentation.
        :param str description: Description of the web server for swagger documentation.
        :param str api_version: API Version of the web server for swagger documentation.
        :param list,str swagger_list: List or str of swagger named services to show in swagger docs.
        :param list,str swagger_list_type: list or str of invokable type can be None, 'sync' or 'async'.

        """

        asset_list = {}
        if swagger_list is None:
            asset_list = invoke_handler.asset_list
        else:
            if isinstance(self._swagger_list, str):
                invoke_item = invoke_handler.find_item(swagger_list)
                if invoke_item:
                    asset_list[invoke_item.asset_id] = invoke_item
            elif isinstance(swagger_list, (tuple, list, set)):
                for name in swagger_list:
                    invoke_item = invoke_handler.find_item(name)
                    if invoke_item:
                        asset_list[invoke_item.asset_id] = invoke_item

        blueprint = get_swagger_blueprint(
            app,
            api_uri,
            invoke_uri,
            asset_list,
            swagger_list_type,
            title,
            description,
            api_version
        )
        app.register_blueprint(blueprint, url_prefix=api_uri)

        doc_ui = get_swaggerui_blueprint(
            doc_uri,
            api_uri,
            config={
                'app_name': title
            },
        )
        app.register_blueprint(doc_ui, url_prefix=doc_uri)

    def add_package(self, package_name):
        """

        Add a python package or a list of python packages to the InvokeHandler.

        :param str,list package_name: Name of the python package(s) that contain the service classes.

        """
        self._invoke_handler.register_package(package_name)

    def add_module(self, module_name):
        """

        Add a python module or a list of python modules to the InvokeHandler.

        :param str,list module_name: Name of the python module(s) that contain the service classes.

        """
        self._invoke_handler.register_module(module_name)

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def title(self):
        return self._title

    @property
    def description(self):
        return self._description

    @property
    def is_swagger(self):
        return self._is_swagger

    @property
    def invoke_handler(self):
        return self._invoke_handler

    @property
    def url(self):
        """

        Return the URL, based on the servers `host` and `port`.

        """
        return generate_url(self._host, self._port)
