"""

Invoke  Service Handler

This call the inovakable services, using sync or async calls

"""

import importlib
import inspect
import logging
import re
import secrets
import time
import types
from threading import (
    RLock,
    Thread
)

import inflection
from yaml import load

from invokable import JobStatus
from invokable.invoke_item import InvokeItem
from invokable.invoke_service import InvokeService
from invokable.job_item import JobItem

logger = logging.getLogger('invokable.invoke_handler')
DEFAULT_JOB_ID_LENGTH = 4
DEFAULT_PURGE_JOB_SECONDS = (60 * 60 * 12)            # auto delete jobs after 12 hours


def invoke_register(func):
    print('inovke register')
    print(func.__doc__)
    return func


class InvokeHandler():

    def __init__(self, job_id_length=None, purge_job_seconds=None):
        """

        Create a basic InovkeHandler object.

        :param int job_id_length: number of hex chars for the job id. Default: 4
        :param purge_job_seconds: number of seconds before deleting completed jobs,
            if <= 0 then do not purge completed jobs. Default 12 hours

        """
        self._invoke_items = {}
        self._result = None
        self._job_items = {}
        self._lock = RLock()

        if purge_job_seconds is None:
            purge_job_seconds = DEFAULT_PURGE_JOB_SECONDS
        self._purge_job_seconds = int(purge_job_seconds)
        if job_id_length is None:
            job_id_length = DEFAULT_JOB_ID_LENGTH
        self._job_id_length = job_id_length

    def register(self, func, invoke_name=None):
        """
        Register a decorated function. using YAML text to define the invokable service.

        :param function func: Function that is passed with the comment text as the definition of the invokable service.
        :param str invoke_name: Invokable name used for this service. If not used then use the 'name' field
            or use the function name. The name will be used in the following order:

            1. parameter invoke_name.
            2. name field in the YAML help text in the function.
            3. function name.

        :returns: The function to execute
        :type: func

        An examlpe of a simple invokable service defined using a function decoration:

        .. code-block:: python

            from invokable-py import InvokeHandler

            handler = InvokeHandler()

            @handler.register
            def hash_text_decorator(service, inputs):
                \"""

                Hash some text using invokbale services

                ---
                name: hash_text_2
                creator: Dex
                modes:
                  - sync
                  - async
                tags:
                  - hash
                  - text
                inputs:
                  to_hash:
                    type: string
                    required: true
                    description: item to hash

                outputs:
                  hash_value:
                    type: string
                    description: return hash value of to_hash

                \"""
                message = hashlib.md5()
                message.update(inputs['to-hash'].encode('utf-8'))
                return {
                    'hash_value': message.hexdigest()
                }

        """
        if isinstance(func, types.FunctionType):
            docstring = func.__doc__
            if invoke_name is None:
                invoke_name = func.__name__
        elif isinstance(func, str):
            docstring = func
        else:
            raise TypeError('you can only pass a function or string')
        items = re.split(r'^\s+---$', docstring, 1, re.MULTILINE)
        if len(items) <= 1:
            raise SyntaxError('Cannot find the "---" seperator')

        info = load(items[1])
        if 'description' in info:
            info['description'] = items[0].strip()
        if 'name' in info:
            invoke_name = info['name']
        if invoke_name is None:
            raise ValueError('you need to give this operation a name')
        info['name'] = invoke_name
        self.register_function(info, func)
        return func

    def register_function(self, name, target, inputs=None, outputs=None, description=None,
                          modes=None, creator=None, license_type=None, date_created=None, in_language=None,
                          operation=None):

        if isinstance(name, dict):
            values = name
            name = values['name']

            field_list = [
                'inputs'
                'outputs',
                'description',
                'modes',
                'creator',
                'license_type',
                'date_created',
                'in_language',
                'operation',
            ]
            data = {
                'name': values['name'],
            }
            for field in field_list:
                if field in values:
                    data[field] = values[field]
        elif isinstance(name, str):
            data = {
                'name': name,
                'inputs': inputs,
                'outputs': outputs,
                'description': description,
                'modes': modes,
                'creator': creator,
                'license_type': license_type,
                'date_created': date_created,
                'in_language': in_language,
                'operation': operation,
            }
        else:
            raise ValueError('You must provide a invokable service name')

        metadata_text = InvokeService.generate_metadata_from_dict(data)
        if name in self._invoke_items:
            logger.warning(f'Found another function or class with the same name {name}')
        invoke_item = InvokeItem(
            name,
            InvokeService,
            metadata_text,
            target
        )
        self._invoke_items[name] = invoke_item
        return [name]

    def register_class(self, invoke_class):
        """

        Register a class to be used as an InvokeService. This class must inherit
        the class InvokeService, and provide static fields defining the input and
        output of the service.

        :param class invoke_class: Class to register for the invoke handler.
        :return: Return the name of the invoke service.
        :type: str

        """
        if isinstance(invoke_class, (tuple, list, set)):
            names = []
            for item in invoke_class:
                names += self.register_class(item)
            return names

        invoke = invoke_class()

        invoke_name = getattr(invoke, 'name', inflection.underscore(invoke.__class__.__name__))
        metadata_text = invoke.generate_metadata()
        if invoke_name in self._invoke_items:
            logger.warning(f'Found another class {invoke_class} with the same name {invoke_name}')

        invoke_item = InvokeItem(
            invoke_name,
            invoke_class,
            metadata_text
        )

        self._invoke_items[invoke_name] = invoke_item
        return [invoke_name]

    def register_package(self, package_name):
        """

        Register a python package for invokable services. This package must
        import all of the classes in the __init__.py file. For this package
        function to work.

        :param str package_name: Name of the package to import
        :return: List of invoke services registered
        :type: list

        """
        # if package name, load in the modules imported in the __init__.py file
        names = []
        if isinstance(package_name, (tuple, list, set)):
            for name in package_name:
                names += self.register_package(name)
            return names
        package_module = importlib.import_module(package_name)
        for item in inspect.getmembers(package_module, inspect.ismodule):
            class_list = InvokeHandler._find_class_list_in_module(item[1])
            names += self.register_class(class_list)
        return names

    def register_module(self, module_name):
        """

        Register a python module for invoke service. This is usually a python module
        file. The module can contain one or many classes.

        :param str module_name: Name of the python module to import.
        :return: List of service names imported
        :type: list

        """
        names = []
        if isinstance(module_name, (tuple, list, set)):
            for name in module_name:
                names += self.register_module(name)
            return names

        # module_name can be an individual module name or package name
        invoke_module = importlib.import_module(module_name)
        class_list = InvokeHandler._find_class_list_in_module(invoke_module)
        return self.register_class(class_list)

    def find_item(self, asset_name):
        """

        Find a service given it's internal name, or asset_id.

        :param str asset_name: Name of the service or the asset_id of the service to find.
        :return: Return the service item, if found. If not found return None
        :type: dict or None.

        """
        for _, invoke_item in self._invoke_items.items():
            # print(asset_name, item['name'], item['asset_id'])
            if invoke_item.name == asset_name or invoke_item.asset_id.lower() == asset_name.lower():
                return invoke_item
        return None

    def invoke(self, name, inputs=None, is_async=False, data=None):
        """

        Invoke a service based on it's service name or asset_id.

        :param str name: Name of the service or asset_id of the service.
        :param dict inputs: Inputs send to the service when the service is started.
        :param boolean is_async: Default to False, if set to True return the job_id instead of the service result.
        :param dict data: Dict of data to send to the invoke service.

        :return: If async, a dict containing the `job_id`, else the result from calling the service.
        :type: dict
        """
        result = {
            'status': JobStatus.FAILED,
            'outputs': {
                'errorcode': 500,
                'description': 'invoke asset_id not found',
            }
        }

        invoke_item = self.find_item(name)
        if invoke_item:
            self._purge_job_list()
            invoke = invoke_item.invoke_class(invoke_item=invoke_item)
            job_id = self._generate_job_id()
            thread = Thread(target=self._invoke_thread, kwargs={
                'job_id': job_id,
                'invoke': invoke,
            })
            self._add_job(job_id, thread, invoke_item, inputs, data)
            thread.start()
            result = {'job-id': job_id}
            if not is_async:
                thread.join()
                result = self.get_job_result(job_id)
        return result

    def get_job(self, job_id):
        """

        Return the current job based on the job_id.

        :param str job_id: Job id returned from the async invoke method.
        :return: JobItem object of the current job
        :type: string

        """
        result = None
        self._lock.acquire()
        if job_id in self._job_items:
            result = self._job_items[job_id]
        self._lock.release()
        return result

    def get_job_status(self, job_id):
        """

        Return the current job status of a job_id.

        :param str job_id: Job id returned from the async invoke method.
        :return: Status string of the current job
        :type: string

        """
        self._lock.acquire()
        result = JobStatus.FAILED
        if job_id in self._job_items:
            result = self._job_items[job_id].status
        self._lock.release()
        return result

    def get_job_result(self, job_id):
        """

        Return the job result of a completed job. If the job is not completed then
        the 'status' will contain the current job status.

        :param str job_id: Job id of the job to get the status.
        :return: Job status
        :type: str

        """
        result = {
            'status': JobStatus.FAILED,
            'outputs': {
                'description': 'Cannot find Job id',
                'errorcode': 501
            }
        }
        if job_id in self._job_items:
            job_item = self._job_items[job_id]
            result = {
                'status': job_item.status,
                'outputs': job_item.outputs
            }
        return result

    def is_job_done(self, job_id):
        """

        Return True if the current job has completed. This function will allways return
        True if the job has failed or has been cancelled.

        :param str job_id: Job id of the job to test
        :return: True if the job has finished, False if the job is still running.
        :type: boolean

        """
        if job_id in self._job_items:
            return self._job_items[job_id].is_done
        return True

    def get_job_list(self):
        """

        Get the list of jobs that have been run and are running.

        :return: return a dict of job_id->status
        :type: dict

        """
        result = {}
        for job_id, job_item in self._job_items.items():
            result[job_id] = job_item.status
        return result

    def unregister(self, name):
        """
        Unregister a service, using it's name.

        :param str name: Name of the service.

        """
        del self._invoke_items[name]

    def get_item(self, name):
        """

        Get a invoke item record, based on the service name.

        :param str name: Name of the service.
        :return: InvokeItem object

        :type: :class:`.InvokeItem`

        """
        if name in self._invoke_items:
            return self._invoke_items[name]
        return None

    @property
    def asset_list(self):
        """

        Return a list of asset_id's for the list of services.

        :return: Dict of asset_id as the key and item as the value.
        :type: dict

        """
        result = {}
        for _, invoke_item in self._invoke_items.items():
            result[invoke_item.asset_id] = invoke_item
        return result

    @property
    def name_list(self):
        return list(self._invoke_items.items())

    @property
    def jobs(self):
        return self._job_items

    @staticmethod
    def _find_class_list_in_module(invoke_module):
        invoke_class_list = []
        for name, obj in inspect.getmembers(invoke_module, inspect.isclass):
            if issubclass(obj, InvokeService) and (name != 'InvokeService'):
                invoke_class_list.append(obj)
        return invoke_class_list

    def _invoke_thread(self, job_id=None, invoke=None):
        if invoke and job_id:
            self._set_job_status(job_id, JobStatus.RUNNING)
            job_item = self.get_job(job_id)
            result = invoke.run(job_item.inputs, job_item.data)
            if result:
                if isinstance(result, (tuple, list, set)):
                    self._set_job_fail(job_id, result[0], result[1])
                else:
                    self._set_job_done(job_id, result)
            else:
                self._set_job_fail(job_id, 'No result returned', 502)
        else:
            self._set_job_fail(job_id, 'Cannot find Job id or invoke', 501)

    def _generate_job_id(self):
        while True:
            job_id = secrets.token_hex(self._job_id_length)
            if job_id not in self._job_items:
                break
        return job_id

    def _add_job(self, job_id, thread, invoke_item, inputs, data=None):
        job_item = JobItem(job_id, thread, invoke_item, JobStatus.SCHEDULE, inputs, data)
        if self._purge_job_seconds > 0:
            job_item.timeout = time.time() + self._purge_job_seconds

        self._lock.acquire()
        self._job_items[job_id] = job_item
        self._lock.release()

    def _purge_job_list(self):
        purge_list = []
        for job_id, job_item in self._job_items.items():
            if job_item.is_timeout:
                if self.is_job_done(job_id):
                    purge_list.append(job_id)

        self._lock.acquire()
        for job_id in purge_list:
            del self._job_items[job_id]
        self._lock.release()

    def _set_job_status(self, job_id, status):
        self._lock.acquire()
        if job_id in self._job_items:
            self._job_items[job_id].status = status
        self._lock.release()

    def _set_job_fail(self, job_id, description, error_code, status=None):
        if status is None:
            status = JobStatus.FAILED
        self._lock.acquire()
        if job_id in self._job_items:
            self._job_items[job_id].outputs = {
                'description': description,
                'errorcode': error_code
            }
            self._job_items[job_id].status = status
        self._lock.release()

    def _set_job_done(self, job_id, result, status=None):
        if status is None:
            status = JobStatus.SUCCEEDED

        self._lock.acquire()
        if job_id in self._job_items:
            self._job_items[job_id].status = status
            self._job_items[job_id].outputs = result
        self._lock.release()
