"""

Service Abstract Base Class


"""

import datetime
import json
import logging
import os
import inflection

from invokable.constants import (
    API_METADATA_VERSION,
    METADATA_ADDITIONAL_INFORMATION_NAME
)

from invokable.metadata_tools import metadata_validator

ALLOWED_DATA_TYPES = ['asset', 'json']

logger = logging.getLogger('invokable.invoke_service')


class InvokeService():

    MODE_ASYNC = 'async'
    MODE_SYNC = 'sync'
    MODES_ALL = [MODE_SYNC, MODE_ASYNC]

    def __init__(self, invoke_item=None):
        """

        When you inherit this service, and you overload the __init__ method, then you will need to call this
        to pass the user data to the calling run method.

        :param object invoke_item: Invoke Item of the invoke call that is being made.

        """
        self._invoke_item = invoke_item

    def run(self, inputs, data=None):
        """


        For each valid Invokable service, you need to overload this method.
        The inputs can be retreived by using self.inputs.
        inputs  - self.invoke_inputs
        args    - self.args
        kwargs  - self.kwargs

        :param dict inputs: Inputs for the invoke call.
        :param dict dta: Optional data to pass to the calling run method.

        :return: tuple error_descripton, error_code or
            dict { 'my_result': 000 }
        :type: tuple or dict
        """
        if self._invoke_item and self._invoke_item.target:
            return self._invoke_item.target(self, inputs, data)

        return None

    def generate_metadata(self):
        """

        Called to generate the metadata based on the static fields defined by the parent class.
        Using the metadata, the asset_id for this service is calculated.

        :return: JSON formated metadata record
        :type: str

        """

        data = {
            'name': getattr(self, 'name', inflection.underscore(self.__class__.__name__)),
            'inputs': getattr(self, 'inputs', None),
            'outputs': getattr(self, 'outputs', None),
            'creator': getattr(self, 'creator', None),
            'date_created': getattr(self, 'date_created', None),
            'license_type': getattr(self, 'license_type', None),
            'in_language': getattr(self, 'in_language', 'en'),
            'description': getattr(self, 'description', None),
            'operation': getattr(self, 'operation', 'operation'),
            'modes': getattr(self, 'modes', InvokeService.MODES_ALL)
        }
        return InvokeService.generate_metadata_from_dict(data)

    @staticmethod
    def generate_metadata_from_dict(data):
        """

        Called to generate the metadata based on the static fields defined by the parent class.
        Using the metadata, the asset_id for this service is calculated.

        :return: JSON formated metadata record
        :type: str

        """

        date_created = data.get('date_created', None)
        if date_created:
            date_created = datetime.datetime.fromisoformat(date_created)
        else:
            module_ctime = os.path.getctime(__file__)
            date_created = datetime.datetime.fromtimestamp(module_ctime)

        operation_modes = data.get('modes', InvokeService.MODES_ALL)
        if isinstance(operation_modes, str):
            operation_modes = [operation_modes]

        metadata = {
            'name': data.get('name', None),
            'license': data.get('license_type', None),
            'creator': data.get('creator', None),
            'dateCreated': date_created.isoformat(),
            'inLanguage': data.get('in_language', 'en'),
            'description': data.get('description', None),
            'type': data.get('type', 'operation'),
            'operation': {
                'modes': operation_modes,
                'inputs': InvokeService._convert_inputs_outputs_to_standard(data.get('inputs', None)),
                'outputs': InvokeService._convert_inputs_outputs_to_standard(data.get('outputs', None)),
            },
            'tags': data.get('tags', []),
            'additionalInformation': {
                METADATA_ADDITIONAL_INFORMATION_NAME: {
                    'version': API_METADATA_VERSION,
                }
            }
        }

        if 'inputs' in data:
            metadata['additionalInformation'][METADATA_ADDITIONAL_INFORMATION_NAME]['inputs'] = data['inputs']

        if 'outputs' in data:
            metadata['additionalInformation'][METADATA_ADDITIONAL_INFORMATION_NAME]['outputs'] = data['outputs']

        delete_names = []
        for name, value in metadata.items():
            if value is None:
                delete_names.append(name)
        for name in delete_names:
            del metadata[name]

        if metadata['operation']['modes'] is None:
            del metadata['operation']['modes']

        metadata_validator(metadata)
        return json.dumps(metadata)

    @staticmethod
    def _convert_inputs_outputs_to_standard(values):
        results = {}
        if values is None:
            return

        if not isinstance(values, dict):
            raise TypeError('The inputs and/or outputs must be defined as dict')

        allowed_values = ['type', 'required', 'position']
        for field_name, item in values.items():
            result = {}
            for name, value in item.items():
                if name in allowed_values:
                    if name == 'type':
                        if value not in ALLOWED_DATA_TYPES:
                            value = 'json'
                    result[name] = value
            results[field_name] = result
        return results

    @property
    def invoke_item(self):
        return self._invoke_item
