"""

 Invoke API

"""

import json
import logging

from flask import (
    Blueprint,
    jsonify,
    request
)

logger = logging.getLogger('invokable.ui.invoke')


def get_invoke_blueprint(uri, handler, authentication, data=None):
    invoke_ui = Blueprint(uri, 'invokable.ui.invoke')

    @invoke_ui.route('/sync/<invoke_name>', methods=['POST'])
    @authentication.login_maybe_required
    def invoke_sync_method(invoke_name):
        """
        Invoke an invokable service, and wait for the return of the service.
        ---
        tags:
          - invoke api
        parameters:
          - name: invoke_name
            in: path
            description: name or asset_id of the invokable service.
            required: true
            type: string
          - name: body
            in: body
            required: true
            schema:
              type: object
        responses:
          200:
            description: return the json object result from the invokable service.
          404:
            description: field value error
          500:
            description: database error
        """
        inputs = json.loads(request.data)

        result = handler.invoke(invoke_name, inputs=inputs, data=data)
        return jsonify(result)

    @invoke_ui.route('/async/<invoke_name>', methods=['POST'])
    @authentication.login_maybe_required
    def invoke_async_method(invoke_name):
        """
        Invoke an invokable service in the background, return a job_id for the job.
        ---
        tags:
          - invoke api
        parameters:
          - name: invoke_name
            in: path
            description: name or asset_id of the invokable service.
            required: true
            type: string
        responses:
          200:
            description: return the job id of the running job.
          404:
            description: field value error
          500:
            description: database error
        """

        inputs = json.loads(request.data)

        result = handler.invoke(invoke_name, inputs=inputs, is_async=True, data=data)
        return jsonify(result)

    @invoke_ui.route('/jobs/<job_id>', methods=['GET'])
    @authentication.login_maybe_required
    def get_job(job_id):
        """
        Get job information based on the job_id
        ---
        tags:
          - invoke api
        parameters:
          - name: job_id
            in: path
            description: job details for the given job_id.
            required: true
            type: string
        responses:
          200:
            description: return a json object of the job
          404:
            description: field value error
          500:
            description: database error
        """
        job_item = handler.get_job(job_id)
        if job_item:
            return jsonify({
                'status': job_item.status,
                'outputs': job_item.outputs
            })
        return 'job not found', 404

    @invoke_ui.route('/jobs/', methods=['GET'])
    @invoke_ui.route('/jobs', methods=['GET'])
    @authentication.login_maybe_required
    def job_list():
        """
        Get job list
        ---
        tags:
          - invoke api
        responses:
          200:
            description: return a list of jobs the jobs status.
          404:
            description: field value error
          500:
            description: database error
        """
        result = {}
        for job_id, job_item in handler.jobs.items():
            result[job_id] = job_item.status
        return jsonify(result)

    return invoke_ui
