"""

 Invoke Metadata API

"""

import logging

from flask import (
    Blueprint,
    Response,
    jsonify,
    request
)
from starfish.utils.crypto_hash import hash_sha3_256

logger = logging.getLogger('invokable.ui.metadata')

STORAGE_METADATA = 'metadata'


def get_metadata_blueprint(uri, handler, storage, authentication):
    meta_data_ui = Blueprint(uri, 'invokable.metadata')

    @meta_data_ui.route('/data/<asset_id>', methods=['GET'])
    @authentication.login_maybe_required
    def get_metadata(asset_id):
        """
        Get metadata based on the asset_id
        ---
        tags:
          - metadata api
        parameters:
          - name: asset_id
            in: path
            description: asset_id to get the metadata
            required: true
            type: string
        responses:
          200:
            description: return a json object of the assets metadata
          404:
            description: field value error
          500:
            description: database error
        """
        item = handler.find_item(asset_id)
        if not item and storage:
            item = storage.read(STORAGE_METADATA, asset_id)

        if item:
            response = Response(
                response=item.metadata_text,
                status=200,
                mimetype='application/json'
            )
            return response
        return 'asset not found', 404

    @meta_data_ui.route('/data', methods=['GET'])
    @authentication.login_maybe_required
    def get_asset_id_list():
        """
        Get metadata asset_id list
        ---
        tags:
          - metadata api
        responses:
          200:
            description: return a list of invokable asset ids and their names.
          404:
            description: field value error
          500:
            description: database error
        """
        result = {}
        for asset_id, item in handler.asset_list.items():
            result[asset_id] = item.name
        return jsonify(result)

    @meta_data_ui.route('/index', methods=['GET'])
    @authentication.login_maybe_required
    def get_data_list():
        """
        Get metadata list dict
        ---
        tags:
          - metadata api
        responses:
          200:
            description: return a list of invokable asset ids and their names.
          404:
            description: field value error
          500:
            description: database error
        """
        result = {}
        for asset_id, item in handler.asset_list.items():
            result[asset_id] = item.metadata
        if storage:
            items = storage.items(STORAGE_METADATA)
            for asset_id, item in items.items():
                result[asset_id] = item
        return jsonify(result)

    @meta_data_ui.route('/data/<asset_id>', methods=['PUT'])
    @authentication.login_maybe_required
    def set_meteadata(asset_id):
        """
        Upload metadata
        ---
        tags:
          - metadata api
        responses:
          200:
            description: return asset_id after a successfull add
          404:
            description: field value error
          500:
            description: database error
        """

        if storage and request.data:
            if not storage.write(STORAGE_METADATA, asset_id, request.data):
                return 'unable to save metadata', 507
        else:
            return 'metadata storage not available', 404

        return asset_id

    @meta_data_ui.route('/data', methods=['POST'])
    @authentication.login_maybe_required
    def add_meteadata():
        """
        add metadata
        ---
        tags:
          - metadata api
        responses:
          200:
            description: return asset_id after a successfull add
          404:
            description: field value error
          500:
            description: database error
        """

        asset_id = ''
        if storage and request.data:
            asset_id = hash_sha3_256(request.data)
            if not storage.write(STORAGE_METADATA, asset_id, request.data):
                return 'unable to save metadata', 507
        else:
            return 'metadata storage not available', 404

        return asset_id

    return meta_data_ui
