"""

 Swagger UI

"""

import logging
from flask import (
    Blueprint,
    jsonify
)
from flask_swagger import swagger

logger = logging.getLogger('invokable.ui.swagger')


def dict_path_exists(path_list, values):
    walk_value = values
    for name in path_list:
        if name not in walk_value:
            return False
        walk_value = walk_value[name]
    return True


def convert_type_swagger_type(name, item):
    swagger_item = {
        'type': item['type'],
        'description': item.get('description', f'{name} parameter')
    }
    if name == 'did':
        swagger_item['x-resolve'] = 'did-asset'

    return swagger_item


def generate_swagger_info(metadata, additonal_data, operation_mode, asset_id, name, is_asset_id=False):
    is_sync = operation_mode == 'sync'
    sync_schema = {
        'type': 'object',
        'description': 'invokable sync result object',
        'properties': {
            'results': {
                'type': 'object',
                'description': 'The invokable service sync result object',
            },
            'status': {
                'type': 'string',
                'description': 'The status result'
            }
        }
    }

    async_schema = {
        'type': 'object',
        'description': 'invokable async result object',
        'properties': {
            'job_id': {
                'type': 'string',
                'description': 'The background job_id to check for completion',
            },
        }

    }
    sub_filter = 'using name'
    if is_asset_id:
        sub_filter = 'using asset id'
    description = metadata.get('description', 'invoke service')
    result = {
        'post': {
            'description': f'{description}',
            'responses': {
                '200': {
                    'description': 'invokable service result object',
                    'schema': sync_schema if is_sync else async_schema
                },
                '404': {'description': 'field value error'},
                '500': {'description': 'database error'}
            },
            'summary': f'{description} ({name})',
            'tags': [f'invoke {operation_mode} services {sub_filter}']
        }
    }
    if additonal_data and 'inputs' in additonal_data:
        param_items = {}
        required_list = []
        input_values = additonal_data['inputs']
        if input_values:
            for field_name, item in input_values.items():
                param_items[field_name] = convert_type_swagger_type(field_name, item)
                if 'required' in item and item['required']:
                    required_list.append(field_name)
            result['parameters'] = [
                {
                    'in': 'body',
                    'name': 'body',
                    'required': True,
                    'schema': {
                        'type': 'object',
                        'required': required_list,
                        'properties': param_items,
                    }
                }
            ]

    if additonal_data and 'outputs' in additonal_data and is_sync:
        output_values = additonal_data['outputs']
        if output_values:
            response_items = {}
            for field_name, item in output_values.items():
                response_items[field_name] = convert_type_swagger_type(field_name, item)
            result['post']['responses']['200']['schema']['properties']['results']['properties'] = response_items
    return result


def get_swagger_blueprint(app, uri, invoke_uri, asset_list, list_type, title, description, api_version):

    if list_type is None:
        list_type = ['sync', 'async']
    elif isinstance(list_type, str):
        list_type = [list_type]

    swagger_ui = Blueprint(uri, 'invokable.swagger')

    @swagger_ui.route('', methods=['GET'])
    def spec():
        swag = swagger(app)
        swag['info']['version'] = api_version
        swag['info']['title'] = title
        swag['info']['description'] = f'{description}`{uri}`.'

        for asset_id, item in asset_list.items():
            metadata = item.metadata
            additonal_data = item.metadata_additional
            name = metadata['name']
            if 'modes' in metadata['operation'] and metadata['operation']['modes']:
                # if metadata['operation']['modes'] in list_type:
                for operation_mode in metadata['operation']['modes']:
                    if operation_mode in list_type:
                        swag['paths'][f'{invoke_uri}/{operation_mode}/{asset_id}'] = \
                            generate_swagger_info(metadata, additonal_data, operation_mode, asset_id, name, True)
                        swag['paths'][f'{invoke_uri}/{operation_mode}/{name}'] = \
                            generate_swagger_info(metadata, additonal_data, operation_mode, asset_id, name)
        return jsonify(swag)
    return swagger_ui
