# Utils

import logging

from starfish import DNetwork
from starfish.account import Account
from starfish.agent import RemoteAgent
from starfish.agent.services import Services

from invokable.constants import DEFAULT_ROUTE_PATHS

logger = logging.getLogger('invokable-py.utils')


def add_to_network(host, port, network_url, address, password, keyfile, route_paths=None):
    """

    Add this server to the Ocean network. This will register the server as an Ocean agent service.

    :param str host: Host name of the server. Default: **127.0.0.1**
    :param int port: Port number of the server. Default: **80**
    :param str netwok_url: URL of the network node to connect too.
    :param str address: Account to preform the registration.
    :param str password: Passwor of the account.
    :param str keyfile: Filename of the account keyfile that contains the encrypted private key.
    :param dict route_paths:  Server route paths to the different api endpoints.
        This can be a single dict name, value or multiple name/values.
        Default: See :py:mod:`invokable.constants`

    :return: DDO object from starfish of the registered server on the Ocean network.
    :type: str

    """
    network = DNetwork(network_url)
    register_account = Account(address, password, keyfile)

    route_paths = load_route_paths(route_paths)
    base_url = generate_url(host, port)

    services = Services(base_url)
    services.add('invoke', route_paths['invoke'])
    if route_paths['meta']:
        services.add('meta', route_paths['meta'])
    ddo = RemoteAgent.generate_ddo(services)
    if network.register_did(register_account, ddo.did, ddo.as_text()):
        logger.info(f'add the invokable server to the network as {ddo.did}')
        return ddo
    return None


def load_route_paths(route_paths=None):
    # set the default paths
    result = DEFAULT_ROUTE_PATHS

    # now add any user specified paths
    if route_paths and isinstance(route_paths, dict):
        for name, value in route_paths.items():
            result[name] = value
    return result


def generate_url(host, port):
    if port == 80:
        return f'http://{host}'
    elif port == 443:
        return f'https://{host}'
    else:
        return f'http://{host}:{port}'
